import logging
from functools import partial

import matplotlib.pyplot as pt
import qcodes as qc

from qblox_instruments import (
        InstrumentClass, InstrumentType,
        )

from .qblox_version import check_qblox_instrument_version, qblox_version, Version
from .q1simulator import Q1Module, run_sequencers


if qblox_version >= Version('0.12'):
    from qblox_instruments import (
        SystemStatuses, SystemStatus, SystemStatusSlotFlags,
        )
    if qblox_version < Version('0.14'):
        from qblox_instruments import SystemState, SystemStatusOld
else:
    from qblox_instruments import (
        SystemState, SystemStatus, SystemStatusSlotFlags
        )


logger = logging.getLogger(__name__)


class ClusterModule(qc.InstrumentChannel, Q1Module):
    def __init__(self, root_instrument, name, slot, n_sequencers=6, sim_type=None):
        super().__init__(root_instrument, name)
        self._slot = slot
        super().init_module(n_sequencers, sim_type)

    def present(self):
        return True

    @property
    def slot_idx(self):
        return self._slot


class EmptySlot(qc.InstrumentChannel):
    def __init__(self, root_instrument, name):
        super().__init__(root_instrument, name)

    def present(self):
        return False


class Cluster(qc.Instrument):
    _cluster_parameters = [
        'trigger_monitor_latest',
        ]
    _log_only_params = [
        'led_brightness',
        ]

    def __init__(self, name, modules={}):
        check_qblox_instrument_version()
        super().__init__(name)

        # TODO return trigger count
        for par_name in self._cluster_parameters:
            self.add_parameter(par_name, set_cmd=partial(self._set, par_name))
        for i in range(1,16):
            par_name = f'trigger{i}_monitor_count'
            self.add_parameter(par_name, set_cmd=partial(self._set, par_name))

        for par_name in self._log_only_params:
            self.add_parameter(par_name,
                               set_cmd=partial(self._log_set, par_name))

        self._modules = {}
        for slot in range(1,21):
            name = f'module{slot}'
            if slot in modules:
                module = ClusterModule(self, name, slot, sim_type=modules[slot])
            else:
                module = EmptySlot(self, name)
            self.add_submodule(name, module)
            self._modules[slot] = module

    def get_idn(self):
        return dict(vendor='Q1Simulator', model='Cluster', serial='', firmware='')

    @property
    def instrument_class(self):
        return InstrumentClass.CLUSTER

    @property
    def instrument_type(self):
        return InstrumentType.MM

    def get_num_system_error(self):
        return 0

    def get_system_error(self):
        return '0,"No error"'

    if qblox_version >= Version('0.12'):
        def get_system_status(self):
            return SystemStatus(
                SystemStatuses.OKAY,
                [],
                SystemStatusSlotFlags({}))

        if qblox_version < Version('0.14'):
            def get_system_state(self):
                return SystemState(
                    SystemStatusOld.OKAY,
                    [],
                    SystemStatusSlotFlags({}))
    else:
        def get_system_state(self):
            return SystemState(
                SystemStatus.OKAY,
                [],
                SystemStatusSlotFlags({}))

    def get_connected_modules(self, filter_fn=None):
        result = {}
        for slot, module in self._modules.items():
            if module.present():
                result[slot] = module
        return result

    def _check_module_present(self, slot):
        if not self._modules[slot].present():
            raise Exception(f'No module in slot {slot}')

    def _set(self, name, value):
        logger.info(f'{self.name}:{name}={value}')

    if qblox_version < Version('0.14'):
        def get_sequencer_state(self, slot, seq_nr, timeout=0):
            return self._modules[slot].get_sequencer_state(seq_nr, timeout)

        def get_acquisition_state(self, slot, seq_nr, timeout=0):
            return self.self._modules[slot].get_acquisition_state(seq_nr, timeout)

    if qblox_version >= Version('0.12'):
        def get_sequencer_status(self, slot, seq_nr, timeout=0):
            return self._modules[slot].get_sequencer_status(seq_nr, timeout)

        def get_acquisition_status(self, slot, seq_nr, timeout=0):
            return self.self._modules[slot].get_acquisition_status(seq_nr, timeout)

    def arm_sequencer(self, slot: int | None = None, sequencer: int | None = None) -> None:
        if slot is not None:
            self._check_module_present(slot)
            self._modules[slot].arm_sequencer(sequencer)
        else:
            for module in self.get_connected_modules().values():
                module.arm_sequencer(sequencer)

    def start_sequencer(self, slot: int | None = None, sequencer: int | None = None) -> None:
        if slot is not None:
            self._check_module_present(slot)
            modules = [self._modules[slot]]
        else:
            modules = self.get_connected_modules().values()

        # Get list of armed sequencers
        # pass to sequence executor
        sequencers = []
        for module in modules:
            seq_numbers = [sequencer] if sequencer is not None else module.armed_seq
            for seq_number in seq_numbers:
                sequencers.append(module.sequencers[seq_number])

        run_sequencers(sequencers)

    def stop_sequencer(self, slot: int | None = None, sequencer: int | None = None) -> None:
        if slot is not None:
            self._check_module_present(slot)
            self._modules[slot].stop_sequencer(sequencer)
        else:
            for module in self.get_connected_modules().values():
                module.stop_sequencer(sequencer)

    @property
    def modules(self):
        return list(self.submodules.values())

    def reset(self):
        self.invalidate_cache()
        for module in self.get_connected_modules().values():
            module.reset()

    def _log_set(self, name, value):
        logger.info(f'{self.name}: {name}={value}')

    def config(self, name, value):
        for module in self.get_connected_modules().values():
            module.config(name, value)

    def get_simulation_end_time(self):
        return max(module.get_simulation_end_time() for module in self.get_connected_modules().values())

    def plot(self,
             t_min: float | None = None,
             t_max: float | None = None,
             channels: list[str] | list[int] | None = None,
             modules: list[int] | None = None,
             create_figure: bool | str = True,
             analogue_filter: bool = False,
             **kwargs):
        """Plots the simulated output of the cluster.

        Args:
            t_min: minimum time in the plot.
            t_max: maximum time in the plot.
            channels: If not None specifies the channels to plot by name or sequencer number.
            modules: If not None specifies the modules to plot by slot number.
            create_figure:
                If True create a new figure.
                If False only pyplot.plot() is called without creating figure or setting axis labels.
                If "modules" creates a new figure per module.
            analogue_filter: plot result after applying (estimated) analog filter.
        """
        if create_figure == True:
            pt.figure()
            pt.title('Cluster')
            pt.grid(True)
            pt.xlabel('[ns]')
            pt.ylabel('[V]')
        for slot, module in self.get_connected_modules().items():
            if modules is not None and slot not in modules:
                # skip module
                continue
            if create_figure == "module":
                pt.figure()
                pt.title(module.label)
                pt.grid(True)
                pt.xlabel('[ns]')
                pt.ylabel('[V]')
            module.plot(t_min=t_min, t_max=t_max, channels=channels, analogue_filter=analogue_filter)
            pt.legend()
        pt.show()



import sys
from dataclasses import dataclass, field
from packaging.version import Version

import matplotlib.pyplot as pt
from qcodes import Instrument

from .q1simulator import Q1Simulator
from .qblox_version import qblox_version


@dataclass
class PlotDef:
    filename: str
    sequencer_name: str | None = None
    out: list[int] = field(default_factory=lambda: [0, 1])
    lo_frequency: float | None = None


def plot_q1asm_file(filename,
                    out=[0, 1],
                    lo_frequency=None,
                    max_render_time=2e6,
                    max_core_cycles=1e7,
                    render_repetitions=False,
                    skip_wait_sync=True,
                    t_min=None,
                    t_max=None,
                    ):
    plot = PlotDef(filename, out=out, lo_frequency=lo_frequency)
    plot_q1asm_files([plot],
                     max_render_time=max_render_time,
                     max_core_cycles=max_core_cycles,
                     render_repetitions=render_repetitions,
                     skip_wait_sync=skip_wait_sync,
                     t_min=t_min,
                     t_max=t_max,
                     )


def plot_q1asm_files(plot_defs,
                     max_render_time=2e6,
                     max_core_cycles=1e7,
                     render_repetitions=False,
                     skip_wait_sync=True,
                     t_min=None,
                     t_max=None,
                     ):

    name = "Q1Viewer"
    try:
        Instrument.find_instrument("Q1Viewer").close()
    except KeyError:
        pass
    sim = Q1Simulator('Q1Viewer', n_sequencers=len(plot_defs), sim_type='Viewer')
    sim.config('max_render_time', max_render_time)
    sim.config('max_core_cycles', max_core_cycles)
    sim.config('skip_loops', () if render_repetitions else ("_start", ))
    sim.config('skip_wait_sync', skip_wait_sync)
    # sim.config('trace', True)
    sim.ignore_triggers = True

    for i, plot in enumerate(plot_defs):
        if plot.sequencer_name:
            sim.sequencers[i].label = plot.sequencer_name

        sequencer = getattr(sim, f'sequencer{i}')
        if plot.lo_frequency is None:
            sequencer.mod_en_awg(False)
        else:
            sequencer.mod_en_awg(True)
            sequencer.nco_freq(plot.lo_frequency)

        for ch in plot.out:
            path = ch % 2
            sequencer.parameters[f'connect_out{ch}'].set('I' if path == 0 else 'Q')

        sequencer.sequence(plot.filename)
        sequencer.sync_en(True)

        sim.arm_sequencer(i)

    sim.start_sequencer()

    for i, plot in enumerate(plot_defs):
        name = plot.sequencer_name if plot.sequencer_name else f'seq{i}'
        if qblox_version < Version("0.12"):
            print(f'State {name}: {sim.get_sequencer_state(i)}')
        else:
            print(f'State {name}: {sim.get_sequencer_status(i)}')

    sim.plot(t_min=t_min, t_max=t_max)
    sim.print_acquisitions()

    sim.close()


def main(argv):
    plot_q1asm_file(argv[0])
    pt.show()


if __name__ == "__main__":
    main(sys.argv[1:])



import logging
import os

from qtpy import QtCore, QtWidgets

from q1simulator.gui.program_details import ProgramDetailsWidget
from q1simulator.gui.qt_utils import (
    qt_log_exception,
    qt_init,
    qt_create_app,
    qt_run_app,
    qt_set_darkstyle,
)


logger = logging.getLogger(__name__)


_app = None


class Q1ProgramBrowser(QtWidgets.QMainWindow):
    """
    Q1Program browser.
    """

    _WINDOW_TITLE: str = "Q1Program Browser"

    def __init__(
        self,
        path: str | None = None,
        gui_style: str | None = None
    ):
        """Creates program browser

        Args:
            path: base directory. If None uses current working directory.
            gui_style: if "dark" uses dark style, otherwise normal style.
        """
        global _app
        logger.debug("Init program browser")

        qt_app_runing = qt_init()
        if not qt_app_runing:
            # note: store reference to avoid garbage collection.
            # reference is also used to restart browser 2nd time in Python console.
            _app = qt_create_app()

        if gui_style == "dark":
            qt_set_darkstyle()

        super().__init__()

        if path is None:
            path = os.getcwd()

        self._path = path

        self.setWindowTitle(self._WINDOW_TITLE)
        self.resize(800, 500)
        self.create_ui()
        self.fill_program_list()
        self.show()

        if _app is not None:
            qt_run_app(_app)

    def create_ui(self):
        content = QtWidgets.QWidget()
        self._list_widget = QtWidgets.QListWidget()
        self._list_widget.setMinimumWidth(260)
        self._list_widget.itemSelectionChanged.connect(self._show_program_info)

        self._program_details = ProgramDetailsWidget(self)

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._list_widget)
        layout.addWidget(self._program_details, 1)

        content.setLayout(layout)
        self.setCentralWidget(content)

    @QtCore.Slot()
    def _show_program_info(self):
        selected_items = self._list_widget.selectedItems()
        if not selected_items:
            return
        item = selected_items[0]
        path = item.data(QtCore.Qt.UserRole)
        self._program_details.show_details(path)

    def fill_program_list(self):
        w = self._list_widget
        w.clear()
        for entry in os.scandir(self._path):
            if not entry.is_dir():
                continue
            i = w.count()
            w.addItem(entry.name)
            item = w.item(i)
            item.setData(QtCore.Qt.UserRole, entry.path)

    def _show_error_message(self, title, message):
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText(message)
        msg.setWindowTitle("Q1ProgramBrowser: " + title)
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        msg.exec_()


if __name__ == "__main__":
    ui = Q1ProgramBrowser(r"C:\measurements\qblox_programs", gui_style="dark")



import logging
from functools import partial
from dataclasses import dataclass
import json
import numpy as np
from typing import Iterator, Iterable

from qcodes.instrument.channel import InstrumentChannel

from .qblox_version import qblox_version, Version

if qblox_version >= Version('0.12'):
    from qblox_instruments import (
        SequencerStatuses,
        SequencerStates,
        )
    if qblox_version < Version('0.14'):
        from qblox_instruments import SequencerStatusOld

if qblox_version < Version('0.14'):
    from qblox_instruments import SequencerState

from qblox_instruments import SequencerStatus, SequencerStatusFlags

from .q1core import Q1Core
from .rt_renderer import Renderer, MockDataEntry


logger = logging.getLogger(__name__)

MockDataType = Iterable[MockDataEntry]


class Q1Sequencer(InstrumentChannel):

    # only logged
    _seq_log_only_parameters = [
        # -- only printed:
        'sync_en',
        'marker_ovr_en',
        'marker_ovr_value',
        'cont_mode_en_awg_path0',
        'cont_mode_en_awg_path1',
        'cont_mode_waveform_idx_awg_path0',
        'cont_mode_waveform_idx_awg_path1',
        'upsample_rate_awg_path0',
        'upsample_rate_awg_path1',
        'nco_freq_cal_type_default',
        ]
    _seq_log_only_parameters_qrm = [
        'connect_acq_I',
        'connect_acq_Q',
        'nco_prop_delay_comp',
        'nco_prop_delay_comp_en',
        ]
    _seq_log_only_parameters_qrm_rf = [
        'connect_acq',
        'nco_prop_delay_comp',
        'nco_prop_delay_comp_en',
        ]

    def __init__(self, parent, name, sim_type):
        super().__init__(parent, name)
        self._is_qcm = sim_type in ['QCM', 'QCM-RF', 'Viewer']
        self._is_qrm = sim_type in ['QRM', 'QRM-RF', 'Viewer']
        self._is_rf = sim_type in ['QCM-RF', 'QRM-RF']

        if self._is_qrm:
            if self._is_rf:
                log_params = self._seq_log_only_parameters + self._seq_log_only_parameters_qrm_rf
            else:
                log_params = self._seq_log_only_parameters + self._seq_log_only_parameters_qrm
            self._v_max = 0.5
        else:
            log_params = self._seq_log_only_parameters
            self._v_max = 2.5

        if sim_type == 'Viewer':
            self._v_max = 2.5

        if self._is_rf:
            self._v_max = 3.3

        for par_name in log_params:
            self.add_parameter(par_name,
                               set_cmd=partial(self._log_set, par_name))

        self.add_parameter('nco_phase_offs', set_cmd=self._nco_phase_offs)
        for i in range(2):
            self.add_parameter(f'gain_awg_path{i}', set_cmd=partial(self._gain_awg, path=i))
            self.add_parameter(f'offset_awg_path{i}', set_cmd=partial(self._offset_awg, path=i))

        self.add_parameter('sequence', set_cmd=self.upload)
        self.add_parameter('mod_en_awg', set_cmd=self._set_mod_en_awg)
        self.add_parameter('nco_freq', set_cmd=self._set_nco_freq)
        self.add_parameter('mixer_corr_gain_ratio', set_cmd=self._set_mixer_gain_ratio)
        self.add_parameter('mixer_corr_phase_offset_degree', set_cmd=self._set_mixer_phase_offset_degree)

        n_out_ch = 4 if self._is_qcm else 2
        if self._is_rf:
            n_out_ch //= 2
        for i in range(n_out_ch):
            self.add_parameter(f'connect_out{i}',
                               set_cmd=partial(self._connect_out, i))

        for i in range(1, 16):
            self.add_parameter(f'trigger{i}_count_threshold',
                               set_cmd=partial(self._set_trigger_count_threshold, i))
            self.add_parameter(f'trigger{i}_threshold_invert',
                               set_cmd=partial(self._set_trigger_threshold_invert, i))

        if self._is_qrm:
            self.add_parameter('demod_en_acq', set_cmd=self._set_demod_en_acq)
            self.add_parameter('integration_length_acq', set_cmd=self._set_integratrion_length_acq)
            self.add_parameter('thresholded_acq_rotation', set_cmd=self._set_thresholded_acq_rotation)
            self.add_parameter('thresholded_acq_threshold', set_cmd=self._set_thresholded_acq_threshold)
            self.add_parameter('thresholded_acq_trigger_en', set_cmd=self._set_thresholded_acq_trigger_en)
            self.add_parameter('thresholded_acq_trigger_address', set_cmd=self._set_thresholded_acq_trigger_address)
            self.add_parameter('thresholded_acq_trigger_invert', set_cmd=self._set_thresholded_acq_trigger_invert)

        self._trace = False
        self.reset()

    def config(self, name, value):
        if name == 'max_render_time':
            self.rt_renderer.max_render_time = value
        elif name == 'max_core_cycles':
            self.q1core.max_core_cycles = value
        elif name == 'trace':
            self._trace = value
            self.rt_renderer.trace_enabled = value
        elif name == 'render_repetitions':
            self.q1core.skip_loops = ("_start", ) if not value else ()
        elif name == 'skip_loops':
            self.q1core.skip_loops = value
        elif name == 'skip_wait_sync':
            self.rt_renderer.skip_wait_sync = value
        elif name == 'acq_trigger_value':
            self.rt_renderer.acq_trigger_value = value

    def reset(self):
        self.waveforms = {}
        self.weights = {}
        self.acquisition_bins = {}
        self._mock_data = {}
        self._trigger_events = []
        self.run_state = 'IDLE'
        self.rt_renderer = Renderer(self.name)
        self.rt_renderer.trace_enabled = self._trace
        self.q1core = Q1Core(self.name, self.rt_renderer, self._is_qrm)
        self.reset_trigger_thresholding()

    def get_simulation_end_time(self):
        return self.rt_renderer.time

    def _log_set(self, name, value):
        logger.info(f'{self.name}: {name}={value}')

    def _nco_phase_offs(self, degrees):
        self.rt_renderer.set_ph((degrees / 360) % 1 * 1e9)

    def _gain_awg(self, gain, path):
        value = int(gain*32767)
        self.rt_renderer.gain_awg_path(value, path)

    def _offset_awg(self, offset, path):
        value = int(offset*32767)
        self.rt_renderer.offset_awg_path(value, path)

    def _set_mod_en_awg(self, value):
        logger.debug(f'{self.name}: mod_en_awg={value}')
        self.rt_renderer.mod_en_awg = value

    def _set_nco_freq(self, value):
        logger.info(f'{self.name}: nco_freq={value}')
        self.rt_renderer.nco_frequency = value

    def _set_demod_en_acq(self, value):
        logger.debug(f'{self.name}: demod_en_acq={value}')
        self.rt_renderer.demod_en_acq = value

    def _set_mixer_gain_ratio(self, value):
        logger.debug(f'{self.name}: mixer_gain_ratio={value}')
        self.rt_renderer.mixer_gain_ratio = value

    def _set_mixer_phase_offset_degree(self, value):
        logger.debug(f'{self.name}: mixer_phase_offset_degree={value}')
        self.rt_renderer.mixer_phase_offset_degree = value

    def _set_channel_map_path_en(self, path, out, enable):
        logger.debug(f'{self.name}: channel_map_path{path}_out{out}_en={enable}')
        if not enable:
            value = 'off'
        elif path == 0:
            value = 'I'
        else:
            value = 'Q'
        self.rt_renderer.connect_out(out, value)

    def _connect_out(self, out, value):
        logger.debug(f'{self.name}: _connect_out{out}={value}')
        self.rt_renderer.connect_out(out, value)

    def _set_trigger_count_threshold(self, address, count):
        self.rt_renderer.set_trigger_count_threshold(address, count)

    def _set_trigger_threshold_invert(self, address, invert: bool):
        self.rt_renderer.set_trigger_threshold_invert(address, invert)

    def _set_integratrion_length_acq(self, value):
        self.rt_renderer.set_integratrion_length_acq(value)

    def _set_thresholded_acq_rotation(self, value):
        self.rt_renderer.set_thresholded_acq_rotation(value)

    def _set_thresholded_acq_threshold(self, value):
        self.rt_renderer.set_thresholded_acq_threshold(value)

    def _set_thresholded_acq_trigger_en(self, value):
        self.rt_renderer.set_thresholded_acq_trigger_en(value)

    def _set_thresholded_acq_trigger_address(self, value):
        self.rt_renderer.set_thresholded_acq_trigger_addr(value)

    def _set_thresholded_acq_trigger_invert(self, value):
        self.rt_renderer.set_thresholded_acq_trigger_invert(value)

    def upload(self, sequence):
        if isinstance(sequence, dict):
            pdict = sequence
        else:
            filename = sequence
            with open(filename) as fp:
                pdict = json.load(fp)
        waveforms = pdict['waveforms']
        weights = pdict['weights']
        acquisitions = pdict['acquisitions']
        program = pdict['program']
        self._set_waveforms(waveforms)
        self._set_weights(weights)
        self._set_acquisition_bins(acquisitions)
        self.q1core.load(program)

    def _set_waveforms(self, waveforms):
        self.waveforms = waveforms
        wavedict = {}
        for name, datadict in waveforms.items():
            index = int(datadict['index'])
            data = np.array(datadict['data'])
            wavedict[index] = data
        self.rt_renderer.set_waveforms(wavedict)

    def _set_weights(self, weights):
        self.weights = weights
        weightsdict = {}
        for name, datadict in weights.items():
            index = int(datadict['index'])
            data = np.array(datadict['data'])
            weightsdict[index] = data
        self.rt_renderer.set_weights(weightsdict)

    def _set_acquisition_bins(self, acq_bins):
        self.acquisition_bins = acq_bins
        bins_dict = {}
        for name, datadict in acq_bins.items():
            index = int(datadict['index'])
            num_bins = int(datadict['num_bins'])
            bins_dict[index] = num_bins
        self.rt_renderer.set_acquisition_bins(bins_dict)

    def _set_rt_mock_data(self):
        for name, md in self._mock_data.items():
            if name not in self.acquisition_bins:
                logger.warning(f"no acquisition_bins for mock_data '{name}'")
                continue
            try:
                data = np.asarray(next(md))
            except StopIteration:
                raise Exception('No more mock data')
            bin_num = int(self.acquisition_bins[name]['index'])
            self.rt_renderer.set_mock_data(bin_num, data)

    if qblox_version < Version('0.14'):
        if qblox_version < Version('0.12'):
            def get_state(self):
                flags = [
                    SequencerStatusFlags[flag.replace(' ', '_')]
                    for flag in self.q1core.errors | self.rt_renderer.errors
                    ]
                if self._is_qrm:
                    flags.append(SequencerStatusFlags.ACQ_BINNING_DONE)
                return SequencerState(
                    SequencerStatus[self.run_state],
                    flags,
                )
        else:
            # deprecated Old version
            def get_state(self):
                flags = [
                    SequencerStatusFlags[flag.replace(' ', '_')]
                    for flag in self.q1core.errors | self.rt_renderer.errors
                    ]
                if self._is_qrm:
                    flags.append(SequencerStatusFlags.ACQ_BINNING_DONE)
                return SequencerState(
                    SequencerStatusOld[self.run_state],
                    flags,
                )

    if qblox_version < Version('0.14'):
        def get_acquisition_state(self):
            if not self._is_qrm:
                raise NotImplementedError('Instrument type is not QRM')
            return True

    if qblox_version >= Version('0.12'):
        def get_status(self):
            info_flags = []
            warn_flags = []
            error_flags = [
                SequencerStatusFlags[flag.replace(' ', '_')]
                for flag in self.q1core.errors | self.rt_renderer.errors
                ]
            log = []
            if self._is_qrm:
                info_flags.append(SequencerStatusFlags.ACQ_BINNING_DONE)

            return SequencerStatus(
                SequencerStatuses.OKAY,
                SequencerStates[self.run_state],
                info_flags,
                warn_flags,
                error_flags,
                log,
            )

    if qblox_version >= Version('0.12'):
        def get_acquisition_status(self):
            if not self._is_qrm:
                raise NotImplementedError('Instrument type is not QRM')
            return True

    def set_trigger_thresholding(self, address: int, count: int, invert: bool) -> None:
        self.rt_renderer.set_trigger_count_threshold(address, count)
        self.rt_renderer.set_trigger_threshold_invert(address, invert)

    def get_trigger_thresholding(self, address: int) -> tuple[int, bool]:
        return (
            self.rt_renderer.get_trigger_count_threshold(address),
            self.rt_renderer.get_trigger_threshold_invert(address)
            )

    def reset_trigger_thresholding(self) -> None:
        for i in range(1, 16):
            self.set_trigger_thresholding(i, 1, False)

    def sideband_cal(self) -> None:
        logger.info(f"Calibrate sideband {self.name}")

    def arm(self):
        self.run_state = 'ARMED'

    def run(self):
        self.run_state = 'RUNNING'
        self.rt_renderer.reset()
        self._set_rt_mock_data()
        self.rt_renderer.trigger_events = self._trigger_events
        self.q1core.run()
        self.run_state = 'STOPPED'

    def get_acquisition_data(self):
        if not self._is_qrm:
            raise NotImplementedError('Instrument type is not QRM')
        cnt, data, thresholded = self.rt_renderer.get_acquisition_data()
        result = {}
        for name, datadict in self.acquisition_bins.items():
            index = int(datadict['index'])
            acq_count = cnt[index]
            path_data = data[index]/acq_count[:, None]
            threshold = thresholded[index]/acq_count

            result[name] = {
                'index': index,
                'acquisition': {
                    'bins': {
                        'integration': {
                            'path0': list(path_data[:, 0]),
                            'path1': list(path_data[:, 1]),
                            },
                        'threshold': threshold,
                        'avg_cnt': list(acq_count),
                    }
                }}
        return result

    def delete_acquisition_data(self, name='', all=False):
        if all:
            self.rt_renderer.delete_acquisition_data_all()
        else:
            index = self.acquisition_bins[name]['index']
            self.rt_renderer.delete_acquisition_data(index)

    # --- Simulator specific methods ---

    def set_acquisition_mock_data(self,
                                  data: Iterable[MockDataType] | None,
                                  name='default',
                                  repeat=False):
        '''
        Sets mock acquisition data for 1 or more runs of the sequence.

        `data` is a list with lists of values to use per run of the sequence.
        The list for a run should have a length equal or bigger than
        the number of acquire calls in the executed sequence.
        The entry for an acquire call is used for both paths.
        If it is a single float value then it is used for both paths.
        If it is a complex value then the real part is used for path 0 and
        the imaginary part for path 1.
        If it is a sequence of two floats then the first is used for path 0
        and the second for path 1.

        Args:
            data:
                if None clears the data and resets default behavior,
                otherwise list of mock data for every run of the sequence.
            name: name of the acquisition
            repeat:
                if True repeatly cycles through the list of mock data,
                otherwise an exception is raised when the list of mock data is exhausted.

        Example:
            # set data for 1 run to return the values 0 till 19 on path 0 and path 1
            data = [np.arange(20)]
            sim.sequencers[0].set_acquisition_mock_data(data)

            # set data for every run to return IQ values with changing phase
            # on path 0 and 1
            data = [np.exp(np.pi*1j*np.arange(20)/10)]
            sim.sequencers[0].set_acquisition_mock_data(data, repeat=True)

            # set data for every run to return the values 0 till 19 on path 0
            # and 100 till 119 on path 1.
            data = [np.arange(20) + 1j*np.arange(100,120)]
            sim.sequencers[0].set_acquisition_mock_data(data, repeat=True)

            # set data for 2 runs to return the values 0 till 19 on the first run
            # and 100 till 119 on the second run.
            data2 = [np.arange(20), np.arange(100, 120)]
            sim.sequencers[0].set_acquisition_mock_data(data2)

            # Return 0...19 and 100...119 alternatingly.
            sim.sequencers[0].set_acquisition_mock_data(data2, repeat=True)

            # reset default behaviour.
            sim.sequencers[0].set_acquisition_mock_data(None)
        '''
        if data is None and name in self._mock_data:
            del self._mock_data[name]
        else:
            self._mock_data[name] = MockData(data, repeat)

    def get_used_triggers(self):
        return self.q1core.get_used_triggers()

    def get_acq_trigger_events(self):
        return self.rt_renderer.acq_trigger_events

    def set_trigger_events(self, events):
        self._trigger_events = events

    def plot(self, t_min=None, t_max=None, analogue_filter=False):
        self.rt_renderer.plot(
            self._v_max,
            plot_label=self.label,
            t_min=t_min,
            t_max=t_max,
            analogue_filter=analogue_filter,
            )

    def get_output(self, t_min=None, t_max=None, analogue_filter=False):
        return self.rt_renderer.get_output(
            self._v_max,
            plot_label=self.label,
            t_min=t_min,
            t_max=t_max,
            analogue_filter=analogue_filter,
            )

    def print_registers(self, reg_nrs=None):
        self.q1core.print_registers(reg_nrs)


@dataclass
class MockData:
    data: Iterable[MockDataType]
    repeat: bool
    data_iter: Iterator[MockDataType] = None

    def __post_init__(self):
        self.data_iter = iter(self.data)

    def __next__(self):
        try:
            return next(self.data_iter)
        except StopIteration:
            if not self.repeat:
                raise
            self.data_iter = iter(self.data)
            return next(self.data_iter)



__version__ = "0.15.3"

from .q1simulator import Q1Simulator
from .cluster import Cluster, ClusterModule
from .q1plotter import Q1Plotter
from .q1viewer import plot_q1asm_file, plot_q1asm_files, PlotDef

from .gui.program_browser import Q1ProgramBrowser



from setuptools import setup, find_packages

packages = ['q1simulator']
print('packages: {}'.format(packages))

setup(name="q1simulator",
	version="0.15.3",
    description="Simulator for Q1ASM",
    author="Sander de Snoo",
	packages = find_packages(),
    python_requires=">=3.10",
	install_requires=[
        'qcodes',
        'qblox_instruments',
        'numpy',
        'matplotlib',
        'qtpy',
        'pyqtgraph',
      ],
    package_data={
        "q1simulator": ["*.hdf5"],
        },
    license='MIT',
	)



import logging
import os
import threading
import traceback

import pyqtgraph as pg
from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QMessageBox

try:
    import IPython.lib.guisupport as gs
    from IPython import get_ipython
except Exception:
    def get_ipython():
        pass


logger = logging.getLogger(__name__)

is_wrapped = threading.local()
is_wrapped.val = False


def qt_log_exception(func):
    ''' Decorator to log exceptions.
    Exceptions are logged and raised again.
    Decorator is designed to be used around functions being called as
    QT event handlers, because QT doesn't report the exceptions.
    Note:
        The decorated method/function cannot be used with
        functools.partial.
    '''

    def wrapped(*args, **kwargs):
        if is_wrapped.val:
            return func(*args, **kwargs)
        else:
            is_wrapped.val = True
            try:
                return func(*args, **kwargs)
            except Exception:
                logger.error('Exception in GUI', exc_info=True)
                raise
            finally:
                is_wrapped.val = False

    return wrapped


def qt_show_exception(message, ex, extra_line: str = None):
    logger.error(message, exc_info=ex)
    text = message
    if extra_line:
        text += "\n" + extra_line
    text += f"\n{type(ex).__name__}: {ex}"
    msg = QMessageBox(
        QMessageBox.Critical,
        "QT-DataViewer: " + message,
        text,
        QMessageBox.Ok,
        )
    msg.setDetailedText("\n".join(traceback.format_exception(ex)))
    msg.setStyleSheet("QTextEdit{min-width:600px}")
    msg.exec_()


def qt_show_error(title, message):
    logger.error(message)
    msg = QMessageBox(
        QMessageBox.Critical,
        "QT-DataViewer: " + title,
        message,
        QMessageBox.Ok,
        )
    msg.exec_()


_qt_app = None


def qt_init(style: str | None = None) -> bool:
    '''Tries to start the QT application if not yet started.
    Most of the cases the QT backend is already started
    by IPython, but sometimes it is not.

    Returns:
        False if QT backend is not running.
    '''
    # application reference must be held in global scope
    global _qt_app

    if _qt_app is None:

        # Set attributes for proper scaling when display scaling is not equal to 100%
        # This should be done before QApplication is started.
        QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
        QtCore.QCoreApplication.setAttribute(QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)

        ipython = get_ipython()

        if ipython:
            if not gs.is_event_loop_running_qt4():
                if any('SPYDER' in name for name in os.environ):
                    raise Exception('Configure QT5 in Spyder -> Preferences -> IPython Console -> Graphics -> Backend')
                else:
                    print('Warning Qt5 not configured for IPython console. Activating it now.')
                    ipython.run_line_magic('gui', 'qt5')

            _qt_app = QtCore.QCoreApplication.instance()
            if _qt_app is None:
                logger.debug('Create Qt application')
                _qt_app = QtWidgets.QApplication([])
            else:
                logger.debug('Qt application already created')
        else:
            _qt_app = QtCore.QCoreApplication.instance()

    if style == "dark":
        qt_set_darkstyle()

    # Only change if still default settings. Do not change is already changed
    if pg.getConfigOption('foreground') == 'd' and pg.getConfigOption('background') == 'k':
        pg.setConfigOption('background', None)
        pg.setConfigOption('foreground', 'k')

    return _qt_app is not None


def qt_set_darkstyle():
    import qdarkstyle
    import pyqtgraph as pg

    qt_app = QtCore.QCoreApplication.instance()
    if qt_app is None:
        return
    dark_stylesheet = qdarkstyle.load_stylesheet()
    # patch qdarkstyle for cropped x-label on 2D graphics.
    dark_stylesheet += r'''
QGraphicsView {
    padding: 0px;
}
'''
    qt_app.setStyleSheet(dark_stylesheet)
    pg.setConfigOption('background', 'k')
    pg.setConfigOption('foreground', 'gray')


def qt_create_app() -> QtCore.QCoreApplication:
    logger.info("Create Qt application")
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)
    app = QtWidgets.QApplication([])
    return app


def qt_run_app(app):
    logger.info("Run Qt Application")
    app.exec()
    logger.info("Qt Application event loop exited")
